/*    
	This file is part of Moose.

    Moose is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as 
    published by the Free Software Foundation, either version 3 of 
    the License, or (at your option) any later version.

    Moose is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public 
    License along with Moose.  If not, see <http://www.gnu.org/licenses/>.
*/

package com.quigley.moose.mapping;

import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.HashMap;
import java.util.Map;

import org.jdom.Element;

public abstract class MappedField {
	public abstract void marshall(Object instance, Mapping mapping, MarshallingContext ctx);
	public abstract void unmarshall(Object instance, Element element, Mapping mapping);
	
	public Class<?> getClazz() {
		return clazz;
	}
	public void setClazz(Class<?> clazz) {
		this.clazz = clazz;
	}
	
	public String getFieldName() {
		return fieldName;
	}
	public void setFieldName(String fieldName) {
		this.fieldName = fieldName;
	}

	public String getXmlName() {
		return xmlName;
	}
	public void setXmlName(String xmlName) {
		this.xmlName = xmlName;
	}
	
	public Object invokeGetter(Object instance) {
		Method getterMethod = getGetter(instance);
		try {
			return getterMethod.invoke(instance, (Object[]) null);
		} catch(Exception e) {
			throw new MappingException("Invocation of method '" + getterMethod.getName() + "' failed.", e);
		}
	}
	protected Method getGetter(Object instance) {
		try {
			return getMethod(instance, getGetterName());
		} catch(MappingException me) {
			return getMethod(instance, getIsName());
		}
	}
	protected String getGetterName() {
		String getterName = "get";
		getterName += fieldName.substring(0, 1).toUpperCase();
		getterName += fieldName.substring(1, fieldName.length());
		return getterName;
	}
	protected String getIsName() {
		String isName = "is";
		isName += fieldName.substring(0, 1).toUpperCase();
		isName += fieldName.substring(1, fieldName.length());
		return isName;
	}
	
	public void invokeSetter(Object instance, Object value) {
		Method setterMethod = getSetter(instance.getClass());
		try {
			setterMethod.invoke(instance, value);
		} catch(Exception e) {
			throw new MappingException("Invocation of method '" + setterMethod.getName() + "' failed.", e);
		}
	}
	protected Method getSetter(Class<?> clazz) {
		Method setterMethod = getMethod(clazz, getSetterName());
		if(setterMethod.getParameterTypes().length == 1) {
			return setterMethod;
		} else {
			throw new MappingException("Setter method '" + setterMethod.getName() + "' takes more than one parameter.");
		}
	}
	protected String getSetterName() {
		String setterName = "set";
		setterName += fieldName.substring(0, 1).toUpperCase();
		setterName += fieldName.substring(1, fieldName.length());
		return setterName;		
	}

	protected Method getMethod(Object instance, String methodName) {
		return getMethod(instance.getClass(), methodName);
	}
	
	protected Method getMethod(Class<?> clazz, String methodName) {
		Map<String, Method> methodIndex = getMethodIndex(clazz.getMethods());
		if(methodIndex.containsKey(methodName)) {
			return methodIndex.get(methodName);
		} else {
			throw new MappingException("Method '" + methodName + "' not available on class '" + clazz.getName() + "'.");
		}
	}

	public Class<?> getFieldType(Class<?> clazz) {
		Method setterMethod = getSetter(clazz);
		Class<?>[] parameterClasses = setterMethod.getParameterTypes();
		return parameterClasses[0];
	}
	
	public Class<?> getParameterizedListType(Class<?> clazz) {
		Type[] typeParameters = getParameterizedTypes(clazz);
		if(typeParameters.length != 1) {
			throw new MappingException("List getter should be parameterized by a single type.");
		}
		return (Class<?>) typeParameters[0];
	}
	
	public Type[] getParameterizedTypes(Class<?> clazz) {
		Map<String, Method> methodIndex = getMethodIndex(clazz.getMethods());
		if(methodIndex.containsKey(getGetterName())) {
			Method getterMethod = methodIndex.get(getGetterName());
			Type getterMethodReturnType = getterMethod.getGenericReturnType();
			if(getterMethodReturnType instanceof ParameterizedType) {
				ParameterizedType parameterizedType = (ParameterizedType) getterMethodReturnType;
				Type[] typeParameters = parameterizedType.getActualTypeArguments();
				return typeParameters;
			}
		} 
		throw new MappingException("Field '" + getFieldName() + "' not parameterized.");
	}		
	
	protected Map<String, Method> getMethodIndex(Method[] methods) {
		Map<String, Method> methodIndex = new HashMap<String, Method>();
		for(Method method : methods) {
			methodIndex.put(method.getName(), method);
		}
		return methodIndex;
	}	
	
	protected Class<?> clazz;
	protected String fieldName;
	protected String xmlName;
}