/*    
	This file is part of Moose.

    Moose is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as 
    published by the Free Software Foundation, either version 3 of 
    the License, or (at your option) any later version.

    Moose is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public 
    License along with Moose.  If not, see <http://www.gnu.org/licenses/>.
*/

package com.quigley.moose.mapping.provider.annotation;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.List;

import com.quigley.moose.handler.BaseTypeHandlerCollection;
import com.quigley.moose.handler.TypeHandlerCollection;
import com.quigley.moose.mapping.MappedArrayField;
import com.quigley.moose.mapping.MappedAttributeField;
import com.quigley.moose.mapping.MappedClass;
import com.quigley.moose.mapping.MappedField;
import com.quigley.moose.mapping.InlineMappedListField;
import com.quigley.moose.mapping.MappedMapField;
import com.quigley.moose.mapping.MappedPropertyField;
import com.quigley.moose.mapping.MappedType;
import com.quigley.moose.mapping.Mapping;
import com.quigley.moose.mapping.MappingException;
import com.quigley.moose.mapping.RootedMappedListField;
import com.quigley.moose.mapping.naming.MappingNameGenerator;
import com.quigley.moose.mapping.provider.MappingProvider;
import com.quigley.moose.mapping.provider.MappingProviderException;
import com.quigley.moose.schema.naming.SchemaNameGenerator;

public class AnnotationMappingProvider implements MappingProvider {
	public AnnotationMappingProvider() {
		//
	}
	public AnnotationMappingProvider(AnnotatedClassesProvider classesProvider) {
		this.annotatedClassesProvider = classesProvider;
	}
	
	public Mapping getMapping() {
		synchronized(AnnotationMappingProvider.class) {
			if(mapping == null) {
				mapping = buildMapping();
			}
		}
		return mapping;
	}

	protected Mapping buildMapping() {
		Mapping m = new Mapping();
		
		if(xmlNamespace != null) {
			m.setXmlNamespace(xmlNamespace);
			m.setXmlPrefix(xmlPrefix);
		}
		
		m.setUseXsiType(useXsiType);
		
		if(mappingNameGenerator != null) {
			m.setMappingNameGenerator(mappingNameGenerator);
		}
		if(schemaNameGenerator != null) {
			m.setSchemaNameGenerator(schemaNameGenerator);
		}
		
		List<MappedType> mappedTypes = new ArrayList<MappedType>();
		if(annotatedClassesProvider != null) {
			for(String className : annotatedClassesProvider.getMappedClassNames()) {
				mappedTypes.add(buildMappedType(className));
			}
		} else {
			throw new MappingProviderException("MappedClassesProvider is null!");
		}
		m.setMappedTypes(mappedTypes);
		
		List<TypeHandlerCollection> typeHandlers = new ArrayList<TypeHandlerCollection>();
		typeHandlers.add(new BaseTypeHandlerCollection());
		m.setTypeHandlerCollections(typeHandlers);
		
		return m;
	}
	
	protected MappedType buildMappedType(String className) {
		try {
			Class<?> clazz = Class.forName(className);
			if(isAnnotatedClass(clazz)) {
				MappedClass mappedClass = new MappedClass();
				mappedClass.setClazz(clazz);

				XML xml = (XML) clazz.getAnnotation(XML.class);
				mappedClass.setMappedParent(getParent(clazz));
				mappedClass.setXmlName(xml.name());
				mappedClass.setMappedFields(buildMappedFieldsForHierarchy(clazz));
			
				return mappedClass;
				
			} else {
				throw new MappingException("No XML annotation found on provided class '" + className + "'.");
			}
			
		} catch(ClassNotFoundException cnfe) {
			throw new MappingProviderException("Unable to get class '" + className + "'.", cnfe);
		}
	}
	
	protected List<MappedField> buildMappedFieldsForHierarchy(Class<?> clazz) {
		List<MappedField> fieldsList = new ArrayList<MappedField>();
		
		Class<?> currentClazz = clazz;
		while(currentClazz != null) {
			if(isAnnotatedClass(currentClazz)) {
				List<MappedField> classFields = buildMappedFields(currentClazz);
				for(MappedField field : classFields) {
					fieldsList.add(field);
				}
			}
			currentClazz = currentClazz.getSuperclass(); 
		}
		
		return fieldsList;
	}
	
	protected Class<?> getParent(Class<?> clazz) {
		Class<?> currentClazz = clazz.getSuperclass();
		while(currentClazz != null) {
			if(isAnnotatedClass(currentClazz)) {
				return currentClazz;
			}
			currentClazz = currentClazz.getSuperclass();
		}
		return null;
	}
	
	protected List<MappedField> buildMappedFields(Class<?> clazz) {
		List<MappedField> mappedFields = new ArrayList<MappedField>();

		for(Field field : clazz.getDeclaredFields()) {
			Annotation[] annotations = field.getAnnotations();
			for(Annotation annotation : annotations) {
				if(annotation instanceof XMLField) {
					XMLField xmlField = (XMLField) annotation;
					if(xmlField.style() == XMLFieldStyle.Node) { 
						MappedPropertyField mappedPropertyField = new MappedPropertyField();
						mappedPropertyField.setFieldName(field.getName());
						mappedPropertyField.setXmlName(xmlField.name());
						mappedPropertyField.setClazz(clazz);
						mappedFields.add(mappedPropertyField);
						
					} else {
						MappedAttributeField mappedAttributeField = new MappedAttributeField();
						mappedAttributeField.setFieldName(field.getName());
						mappedAttributeField.setXmlName(xmlField.name());
						mappedAttributeField.setClazz(clazz);
						mappedFields.add(mappedAttributeField);
					}
					break;
				}
				if(annotation instanceof XMLList) {
					XMLList xmlList = (XMLList) annotation;
					if(xmlList.style() == XMLListStyle.INLINE) {
						InlineMappedListField mappedListField = new InlineMappedListField();
						mappedListField.setFieldName(field.getName());
						mappedListField.setXmlName(xmlList.name());
						mappedListField.setClazz(clazz);
						mappedFields.add(mappedListField);
					} else
					if(xmlList.style() == XMLListStyle.ROOTED) {
						RootedMappedListField mappedListField = new RootedMappedListField();
						mappedListField.setElementXmlName(xmlList.elementName());
						mappedListField.setFieldName(field.getName());
						mappedListField.setXmlName(xmlList.name());
						mappedListField.setClazz(clazz);
						mappedFields.add(mappedListField);
					}
					break;
				}
				if(annotation instanceof XMLMap) {
					XMLMap xmlMap = (XMLMap) annotation;
					MappedMapField mappedMapField = new MappedMapField();
					mappedMapField.setFieldName(field.getName());
					mappedMapField.setKeyXmlName(xmlMap.keyName());
					mappedMapField.setPairXmlName(xmlMap.pairName());
					mappedMapField.setValueXmlName(xmlMap.valueName());
					mappedMapField.setXmlName(xmlMap.name());
					mappedMapField.setClazz(clazz);
					mappedFields.add(mappedMapField);
					break;
				}
				if(annotation instanceof XMLArray) {
					XMLArray xmlArray = (XMLArray) annotation;
					MappedArrayField mappedArrayField = new MappedArrayField();
					mappedArrayField.setElementXmlName(xmlArray.elementName());
					mappedArrayField.setFieldName(field.getName());
					mappedArrayField.setXmlName(xmlArray.name());
					mappedArrayField.setClazz(clazz);
					mappedFields.add(mappedArrayField);
					break;
				}
				if(annotation instanceof XMLCustomField) {
					XMLCustomField xmlCustomField = (XMLCustomField) annotation;
					try {
						MappedField mappedField = (MappedField) xmlCustomField.clazz().newInstance();
						mappedField.setFieldName(field.getName());
						mappedField.setXmlName(xmlCustomField.name());
						mappedField.setClazz(clazz);
						mappedFields.add(mappedField);
						break;
						
					} catch(Exception e) {
						throw new MappingProviderException("Unable to instantiate custom mapped field of type '" + xmlCustomField.clazz().getName() + "'.", e);
					}
				}
			}
		}
		
		return mappedFields;
	}
	
	protected boolean isAnnotatedClass(Class<?> clazz) {
		return (clazz.getAnnotation(XML.class) != null);
	}
	
	public String getXmlNamespace() {
		return xmlNamespace;
	}
	public void setXmlNamespace(String xmlNamespace) {
		this.xmlNamespace = xmlNamespace;
	}

	public String getXmlPrefix() {
		return xmlPrefix;
	}
	public void setXmlPrefix(String xmlPrefix) {
		this.xmlPrefix = xmlPrefix;
	}
	
	public boolean isUseXsiType() {
		return useXsiType;
	}
	public void setUseXsiType(boolean useXsiType) {
		this.useXsiType = useXsiType;
	}

	public MappingNameGenerator getMappingNameGenerator() {
		return mappingNameGenerator;
	}
	public void setMappingNameGenerator(MappingNameGenerator mappingNameGenerator) {
		this.mappingNameGenerator = mappingNameGenerator;
	}
	
	public SchemaNameGenerator getSchemaNameGenerator() {
		return schemaNameGenerator;
	}
	public void setSchemaNameGenerator(SchemaNameGenerator schemaNameGenerator) {
		this.schemaNameGenerator = schemaNameGenerator;
	}

	public AnnotatedClassesProvider getAnnotatedClassesProvider() {
		return annotatedClassesProvider;
	}
	public void setAnnotatedClassesProvider(AnnotatedClassesProvider annotatedClassesProvider) {
		this.annotatedClassesProvider = annotatedClassesProvider;
	}

	protected Mapping mapping;
	
	protected String xmlNamespace;
	protected String xmlPrefix;
	protected boolean useXsiType;
	
	protected MappingNameGenerator mappingNameGenerator;
	protected SchemaNameGenerator schemaNameGenerator;
	protected AnnotatedClassesProvider annotatedClassesProvider;
}