/*
    This file is part of Moose.

    Moose is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as 
    published by the Free Software Foundation, either version 3 of 
    the License, or (at your option) any later version.

    Moose is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public 
    License along with Moose.  If not, see <http://www.gnu.org/licenses/>.
*/

package com.quigley.moose.schema;

import java.lang.reflect.Type;
import java.util.Collections;

import com.quigley.moose.handler.EnumHandler;
import com.quigley.moose.handler.TypeHandler;
import com.quigley.moose.mapping.InlineMappedListField;
import com.quigley.moose.mapping.MappedAttributeField;
import com.quigley.moose.mapping.MappedClass;
import com.quigley.moose.mapping.MappedCustomField;
import com.quigley.moose.mapping.MappedField;
import com.quigley.moose.mapping.MappedMapField;
import com.quigley.moose.mapping.MappedPropertyField;
import com.quigley.moose.mapping.MappedType;
import com.quigley.moose.mapping.Mapping;
import com.quigley.moose.mapping.RootedMappedListField;
import com.quigley.moose.schema.naming.SchemaNameGenerator;

public class SchemaGenerator {
	public static String generate(Mapping mapping) {
		SchemaGeneratorContext ctx = new SchemaGeneratorContext(mapping);

		generateBeginning(ctx);
		
		ctx.indent();
		
		if(ctx.getMappedClasses().size() > 0) {
			ctx.append("<!-- Mapped Classes -->\n\n");
		}
		for(Class<?> mappedClass : ctx.getMappedClasses()) {
			generateMappedClass(mappedClass, ctx);
		}

		if(ctx.getReferencedEnums().size() > 0) {
			ctx.append("<!-- Mapped Enums -->\n\n");
		}
		Collections.sort(ctx.getReferencedEnums(), new ClassComparator());
		for(Class<?> referencedEnum : ctx.getReferencedEnums()) {
			generateReferencedEnum(referencedEnum, ctx);
		}
		
		if(ctx.getListContainers().size() > 0) {
			ctx.append("<!-- Mapped Lists -->\n\n");
		}
		for(ListContainer listContainer : ctx.getListContainers()) {
			generateListContainer(listContainer, ctx);
		}
		
		if(ctx.getMapContainers().size() > 0) {
			ctx.append("<!-- Mapped Maps -->\n\n");
		}
		for(MapContainer mapContainer : ctx.getMapContainers()) {
			generateMapContainer(mapContainer, ctx);
		}
		
		ctx.outdent();
		
		generateEnd(mapping, ctx);
		
		return ctx.getOut().toString();
	}
	
	protected static void generateBeginning(SchemaGeneratorContext ctx) {
		ctx.getOut().append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
		ctx.getOut().append("<xs:schema xmlns:xs=\"http://www.w3.org/2001/XMLSchema\"");
		if(ctx.getMapping().getXmlNamespace() != null) {
			ctx.indent();
			
			ctx.getOut().append("\n");
			if(ctx.getMapping().getXmlPrefix() != null) {
				ctx.append("xmlns:" + ctx.getMapping().getXmlPrefix() + "=\"" + ctx.getMapping().getXmlNamespace() + "\"\n");
			} else {
				ctx.append("xmlns=\"" + ctx.getMapping().getXmlNamespace() + "\"\n");
			}
			ctx.append("targetNamespace=\"" + ctx.getMapping().getXmlNamespace() + "\"\n");
			ctx.append("elementFormDefault=\"qualified\">\n\n");
			
			ctx.outdent();
		} else {
			ctx.getOut().append(">\n\n");
		}
	}
	
	protected static void generateMappedClass(Class<?> clazz, SchemaGeneratorContext ctx) {
		MappedType mappedType = ctx.getMapping().getMappedType(clazz);
		if(mappedType instanceof MappedClass) {
			MappedClass mappedClass = (MappedClass) mappedType;
			SchemaNameGenerator sng = ctx.getMapping().getSchemaNameGenerator();
			
			ctx.append("<xs:element name=\"" + mappedClass.getXmlName() + "\" type=\"" + ctx.getXmlPrefix() + sng.xmlNameToTypeName(mappedClass.getXmlName()) + "\"/>\n");
			ctx.append("<xs:complexType name=\"" + sng.xmlNameToTypeName(mappedClass.getXmlName()) + "\">\n");
			
			MappedClass mappedParentClass = null;
			if(mappedClass.getMappedParent() != null) {
				mappedParentClass = (MappedClass) ctx.getMapping().getMappedType(mappedClass.getMappedParent());
				ctx.indent();
				ctx.append("<xs:complexContent>\n");
				ctx.indent();
				ctx.append("<xs:extension base=\"" + ctx.getXmlPrefix() + sng.xmlNameToTypeName(mappedParentClass.getXmlName()) + "\">\n");
			}
			ctx.indent();
			generateMappedElementFields(mappedClass, ctx);
			generateMappedAttributeFields(mappedClass, ctx);
			ctx.outdent();
			if(mappedClass.getMappedParent() != null) {
				ctx.append("</xs:extension>\n");
				ctx.outdent();
				ctx.append("</xs:complexContent>\n");
				ctx.outdent();
			}
			ctx.append("</xs:complexType>\n\n");
		}
	}
	
	protected static void generateMappedElementFields(MappedClass mappedClass, SchemaGeneratorContext ctx) {
		if(mappedClass.getMappedElementFields() != null && mappedClass.getMappedElementFields().size() > 0) {
			ctx.append("<xs:sequence>\n");
			ctx.indent();
		}
		for(MappedField mappedField : mappedClass.getMappedElementFields()) {
			if(mappedField instanceof MappedPropertyField && mappedField.getClazz() == mappedClass.getClazz()) {
				MappedPropertyField mappedPropertyField = (MappedPropertyField) mappedField;
				SchemaNameGenerator sng = ctx.getMapping().getSchemaNameGenerator();
				
				Class<?> fieldClass = mappedPropertyField.getFieldType(mappedClass.getClazz());
				if(fieldClass.isEnum()) {
					if(!ctx.getReferencedEnums().contains(fieldClass)) {
						ctx.getReferencedEnums().add(fieldClass);
					}
				}
				if(ctx.getMapping().hasTypeHandler(fieldClass)) {
					TypeHandler typeHandler = ctx.getMapping().getTypeHandler(fieldClass);
					ctx.append("<xs:element name=\"" + mappedPropertyField.getXmlName() + "\" type=\"" + typeHandler.getXsdType(ctx.getMapping().getXmlPrefix()) + "\"/>\n");
					
				} else {
					MappedClass fieldMappedType = (MappedClass) ctx.getMapping().getMappedType(fieldClass);
					String fieldMappedTypeName = ctx.getXmlPrefix() + sng.xmlNameToTypeName(fieldMappedType.getXmlName());
					
					ctx.append("<xs:element name=\"" + mappedPropertyField.getXmlName() + "\" type=\"" + fieldMappedTypeName + "\"/>\n");
				}
			} else
			if(mappedField instanceof MappedPropertyField && mappedField.getClazz() != mappedClass.getClazz()) {
				// We can safely ignore this, as this field is defined in our parent class (or elsewhere in the hierarchy).
				
			} else
			if(mappedField instanceof RootedMappedListField && mappedField.getClazz() == mappedClass.getClazz()) {
				RootedMappedListField mappedListField = (RootedMappedListField) mappedField;
				SchemaNameGenerator sng = ctx.getMapping().getSchemaNameGenerator();
				
				Class<?> fieldType = mappedListField.getParameterizedListType(mappedClass.getClazz());
				if(fieldType.isEnum()) {
					if(!ctx.getReferencedEnums().contains(fieldType)) {
						ctx.getReferencedEnums().add(fieldType);
					}
				}
				
				ListContainer listContainer = new ListContainer();
				listContainer.setElementClass(fieldType);
				listContainer.setElementXmlName(mappedListField.getElementXmlName());
				listContainer.setName(sng.xmlNameToListTypeName(mappedListField.getXmlName()));
				ctx.addListContainer(listContainer);
				
				ctx.append("<xs:element name=\"" + mappedListField.getXmlName() + "\" type=\"" + ctx.getXmlPrefix() + sng.xmlNameToListTypeName(mappedListField.getXmlName()) + "\"/>\n");
			} else
			if(mappedField instanceof InlineMappedListField && mappedField.getClazz() == mappedClass.getClazz()) {
				InlineMappedListField mappedListField = (InlineMappedListField) mappedField;
				
				Class<?> listClass = mappedListField.getParameterizedListType(mappedListField.getClazz());
				
				if(listClass.isEnum()) {
					if(!ctx.getReferencedEnums().contains(listClass)) {
						ctx.getReferencedEnums().add(listClass);
					}
				}
				if(ctx.getMapping().hasTypeHandler(listClass)) {
					TypeHandler typeHandler = ctx.getMapping().getTypeHandler(listClass);
					
					ctx.append("<xs:element name=\"" + mappedListField.getXmlName() + "\" type=\"" + typeHandler.getXsdType(ctx.getMapping().getXmlPrefix()) + "\" " + 
							"minOccurs=\"0\" maxOccurs=\"unbounded\"/>\n");
				} else {
					MappedClass listMappedType = (MappedClass) ctx.getMapping().getMappedType(listClass);
					SchemaNameGenerator sng = ctx.getMapping().getSchemaNameGenerator();
					String listMappedTypeName = ctx.getXmlPrefix() + sng.xmlNameToTypeName(listMappedType.getXmlName());
					
					ctx.append("<xs:element name=\"" + mappedListField.getXmlName() + "\" type=\"" + listMappedTypeName + "\" minOccurs=\"0\" maxOccurs=\"unbounded\"/>\n");
				}
			} else
			if(mappedField instanceof MappedMapField && mappedField.getClazz() == mappedClass.getClazz()) {
				MappedMapField mappedMapField = (MappedMapField) mappedField;
				SchemaNameGenerator sng = ctx.getMapping().getSchemaNameGenerator();
				
				Type[] types = mappedMapField.getParameterizedTypes(mappedClass.getClazz());
				Class<?> mapKeyType = (Class<?>) types[0];
				Class<?> mapValueType = (Class<?>) types[1];
				
				if(mapKeyType.isEnum()) {
					if(!ctx.getReferencedEnums().contains(mapKeyType)) {
						ctx.getReferencedEnums().add(mapKeyType);
					}
				}
				if(mapValueType.isEnum()) {
					if(!ctx.getReferencedEnums().contains(mapValueType)) {
						ctx.getReferencedEnums().add(mapValueType);
					}
				}
				
				MapContainer mapContainer = new MapContainer();
				mapContainer.setName(mappedMapField.getPairXmlName());
				mapContainer.setKeyClass(mapKeyType);
				mapContainer.setKeyXmlName(mappedMapField.getKeyXmlName());
				mapContainer.setValueClass(mapValueType);
				mapContainer.setValueXmlName(mappedMapField.getValueXmlName());
				ctx.addMapContainer(mapContainer);
				
				ctx.append("<xs:element name=\"" + mappedMapField.getXmlName() + "\" type=\"" + ctx.getXmlPrefix() + sng.xmlNameToMapEntryTypeName(mappedMapField.getPairXmlName()) + "\"/>\n");
			} else
			if(mappedField instanceof MappedCustomField) {
				ctx.append("<xs:element name=\"" + mappedField.getXmlName() + "\" type=\"" + ((MappedCustomField) mappedField).getXsdType(ctx.getXmlPrefix()) + "\"/>\n");
				
			} else {
				throw new SchemaGeneratorException("I don't know how to handle elements of type '" + mappedField.getClass().getName() + "' (class name='" + mappedClass.getXmlName() + "', field name='" + mappedField.getXmlName() + "')!");
			}
		}
		if(mappedClass.getMappedElementFields() != null && mappedClass.getMappedElementFields().size() > 0) {
			ctx.outdent();
			ctx.append("</xs:sequence>\n");
		}
	}
	
	protected static void generateMappedAttributeFields(MappedClass mappedClass, SchemaGeneratorContext ctx) {
		for(MappedAttributeField mappedAttributeField : mappedClass.getMappedAttributeFields()) {
			Class<?> fieldType = mappedAttributeField.getFieldType(mappedClass.getClazz());
			if(fieldType.isEnum()) {
				if(!ctx.getReferencedEnums().contains(fieldType)) {
					ctx.getReferencedEnums().add(fieldType);
				}
			}
			TypeHandler typeHandler = ctx.getMapping().getTypeHandler(fieldType);
			ctx.append("<xs:attribute name=\"" + mappedAttributeField.getXmlName() + "\" type=\"" + typeHandler.getXsdType(ctx.getMapping().getXmlPrefix()) + "\"/>\n");
		}
	}
	
	protected static void generateReferencedEnum(Class<?> referencedEnum, SchemaGeneratorContext ctx) {
		EnumHandler enumHandler = (EnumHandler) ctx.getMapping().getTypeHandler(referencedEnum);
		
		ctx.append("<xs:simpleType name=\"").append(enumHandler.getEnumXmlName(null)).append("\">\n");
		ctx.indent();
		ctx.append("<xs:restriction base=\"xs:string\">\n");
		
		ctx.indent();
		for(Object value : referencedEnum.getEnumConstants()) {
			Enum<?> enumValue = (Enum<?>) value;
			ctx.append("<xs:enumeration value=\"").append(enumValue.toString()).append("\"/>\n");
		}
		ctx.outdent();
		
		ctx.append("</xs:restriction>\n");
		ctx.outdent();
		ctx.append("</xs:simpleType>\n\n");
	}
	
	protected static void generateListContainer(ListContainer listContainer, SchemaGeneratorContext ctx) {
		SchemaNameGenerator sng = ctx.getMapping().getSchemaNameGenerator();
		
		ctx.append("<xs:complexType name=\"" + listContainer.getName() + "\">\n");

		ctx.indent();
		ctx.append("<xs:sequence>\n");
		
		ctx.indent();
		if(ctx.getMapping().hasMappedType(listContainer.getElementClass())) {
			MappedClass mappedClass = (MappedClass) ctx.getMapping().getMappedType(listContainer.getElementClass());
			ctx.append("<xs:element name=\"" + listContainer.getElementXmlName() + "\" type=\"" + ctx.getXmlPrefix() + sng.xmlNameToTypeName(mappedClass.getXmlName()) + "\" maxOccurs=\"unbounded\"/>\n");
		} else {
			TypeHandler typeHandler = ctx.getMapping().getTypeHandler(listContainer.getElementClass());
			ctx.append("<xs:element name=\"" + listContainer.getElementXmlName() + "\" type=\"" + typeHandler.getXsdType(ctx.getMapping().getXmlPrefix()) + "\" maxOccurs=\"unbounded\"/>\n");
		}
		ctx.outdent();
		
		ctx.append("</xs:sequence>\n");
		ctx.outdent();

		ctx.append("</xs:complexType>\n\n");
	}
	
	protected static void generateMapContainer(MapContainer mapContainer, SchemaGeneratorContext ctx) {
		SchemaNameGenerator sng = ctx.getMapping().getSchemaNameGenerator();
		
		ctx.append("<xs:complexType name=\"" + sng.xmlNameToMapEntryTypeName(mapContainer.getName()) + "\">\n");
		ctx.indent();
		ctx.append("<xs:sequence>\n");
		ctx.indent();
		ctx.append("<xs:element name=\"" + mapContainer.getName() + "\" type=\"" + ctx.getXmlPrefix() + sng.xmlNameToMapPairTypeName(mapContainer.getName()) + "\" maxOccurs=\"unbounded\"/>\n");
		ctx.outdent();
		ctx.append("</xs:sequence>\n");
		ctx.outdent();
		ctx.append("</xs:complexType>\n\n");
		
		ctx.append("<xs:complexType name=\"" + sng.xmlNameToMapPairTypeName(mapContainer.getName()) + "\">\n");
		ctx.indent();
		ctx.append("<xs:sequence>\n");
		ctx.indent();
		if(ctx.getMapping().hasMappedType(mapContainer.getKeyClass())) {
			MappedClass mappedClass = (MappedClass) ctx.getMapping().getMappedType(mapContainer.getKeyClass());
			ctx.append("<xs:element name=\"" + mapContainer.getKeyXmlName() + "\" type=\"" + ctx.getXmlPrefix() + sng.xmlNameToTypeName(mappedClass.getXmlName()) + "\"/>\n");
		} else {
			TypeHandler typeHandler = ctx.getMapping().getTypeHandler(mapContainer.getKeyClass());
			ctx.append("<xs:element name=\"" + mapContainer.getKeyXmlName() + "\" type=\"" + typeHandler.getXsdType(ctx.getMapping().getXmlPrefix()) + "\"/>\n");
		}
		
		if(ctx.getMapping().hasMappedType(mapContainer.getValueClass())) {
			MappedClass mappedClass = (MappedClass) ctx.getMapping().getMappedType(mapContainer.getValueClass());
			ctx.append("<xs:element name=\"" + mapContainer.getValueXmlName() + "\" type=\"" + ctx.getXmlPrefix() + sng.xmlNameToTypeName(mappedClass.getXmlName()) + "\"/>\n");
		} else {
			TypeHandler typeHandler = ctx.getMapping().getTypeHandler(mapContainer.getValueClass());
			ctx.append("<xs:element name=\"" + mapContainer.getValueXmlName() + "\" type=\"" + typeHandler.getXsdType(ctx.getMapping().getXmlPrefix()) + "\"/>\n");
		}
		ctx.outdent();
		ctx.append("</xs:sequence>\n");
		ctx.outdent();
		ctx.append("</xs:complexType>\n\n");
	}
	
	protected static void generateEnd(Mapping mapping, SchemaGeneratorContext ctx) {
		ctx.append("</xs:schema>\n");
	}
}
